/*
 * Simple netcat.
 *
 * snetcat <host> <port>
 *
 * Example:
 * snetcat 127.0.0.1 10000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifndef WIN32
#include <signal.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif
#else
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

static void help()
{
  printf("snetcat <host> <port>\n");
  exit(0);
}

static void error_exit(char *message)
{
  fprintf(stderr, "%s", message);
  exit(1);
}

#ifdef WIN32
int winsock_init()
{
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    error_exit("WSAStartup() failed.\n");
  return 0;
}
#endif

int send_to_socket(int s)
{
  int size, offset, r;
  char buffer[1024*4];

  while (1) {
    /* ɸϤǡɤ߹ */
    r = read(0, buffer, sizeof(buffer));
    if (r <= 0)
      break;

    size = r;
    offset = 0;

    /* ǡ */
    while (size > 0) {
#ifndef WIN32
      r = write(s, buffer + offset, size);
#else
      r = send(s, buffer + offset, size, 0);
#endif
      if (r <= 0)
	break;
      size -= r;
      offset += r;
    }
  }

  return r;
}

int recv_from_socket(int s)
{
  int size, offset, r;
  char buffer[1024*4];

  while (1) {
    /* ǡμ */
#ifndef WIN32
    r = read(s, buffer, sizeof(buffer));
#else
    r = recv(s, buffer, sizeof(buffer), 0);
#endif
    if (r <= 0)
      break;

    size = r;
    offset = 0;

    /* ɸϤ˥ǡ񤭽Ф */
    while (size > 0) {
      r = write(1, buffer + offset, size);
      if (r <= 0)
	break;
      size -= r;
      offset += r;
    }
  }

  return r;
}

#ifndef WIN32
static void sigchld_handler(int value)
{
  int status;
  wait(&status);
}

int create_child(int s)
{
  int pid;

  pid = fork();
  if (pid < 0)
    error_exit("Cannot fork.\n");
  if (pid == 0) {
    send_to_socket(s);
    shutdown(s, SHUT_WR);
    _exit(0);
  }
  signal(SIGCHLD, sigchld_handler);

  return 0;
}
#else
DWORD WINAPI ThreadProc(LPVOID lpParamater)
{
  int s = (int)lpParamater;
  send_to_socket(s);
  shutdown(s, SD_SEND);
  return 0;
}

int create_child(int s)
{
  HANDLE hThread;
  DWORD dwThreadId;

  hThread = CreateThread(NULL, 0, ThreadProc, (LPVOID)s, 0, &dwThreadId);
  if (hThread == 0)
    error_exit("Cannot create thread.\n");

  return 0;
}
#endif

int main(int argc, char *argv[])
{
  char *hostname;
  int port, s;
  struct sockaddr_in addr;
  socklen_t addrlen;
#ifdef USE_GETHOSTBYNAME
  struct hostent *host;
#endif

  argc--; argv++;
  if (argc < 2)
    help();

  hostname = *argv;
  argc--; argv++;
  port = atoi(*argv);
  argc--; argv++;

#ifdef WIN32
  winsock_init();
  if (!isatty(0)) setmode(0, O_BINARY);
#endif

#ifdef USE_GETHOSTBYNAME
  /* ۥ̾饢ɥ쥹 */
  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("Cannot resolve hostname.\n");
#endif

  /* åȤΥץ */
  s = socket(PF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("Cannot open socket.\n");

  /* ɥ쥹 */
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
#ifdef USE_GETHOSTBYNAME
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);
#else
  addr.sin_addr.s_addr = inet_addr(hostname);
#endif
  addrlen = sizeof(addr);

  /* Ф³ */
  if (connect(s, (struct sockaddr *)&addr, addrlen) < 0)
    error_exit("Cannot connect.\n");

  create_child(s);

  recv_from_socket(s);

  /* åȤΥ */
#ifndef WIN32
  shutdown(s, SHUT_RDWR);
  close(s);
#else
  shutdown(s, SD_BOTH);
  closesocket(s);
#endif

  return 0;
}
